
/*
 * pbrt source code Copyright(c) 1998-2005 Matt Pharr and Greg Humphreys
 *
 * All Rights Reserved.
 * For educational use only; commercial use expressly forbidden.
 * NO WARRANTY, express or implied, for this software.
 * (See file License.txt for complete license)
 */

// directlighting.cpp*
#include "pbrt.h"
#include "transport.h"
#include "scene.h"
#include "sunsky.dat"

const float Alpha_1	= 0.83331810215394e-3;
const float Alpha_2	= 0.11360016092149e-3;
#define radians(x) ((x)/180.0*M_PI)

const struct spectrum41_t
{
	int start_wavelength;
	int end_wavelength;
	int nsamples;

	float samples[41];
	float x, y, z;
	float r, g, b;
};

// DirectLighting Declarations
enum LightStrategy { SAMPLE_ALL_UNIFORM, SAMPLE_ONE_UNIFORM,
	SAMPLE_ONE_WEIGHTED  // NOBOOK
};
class DirectLighting : public SurfaceIntegrator {
public:
	// DirectLighting Public Methods
	DirectLighting(LightStrategy ls, int md, const float &turb);
	~DirectLighting();
	Spectrum Li(const Scene *scene, const RayDifferential &ray, const Sample *sample,
		float *alpha) const;
	void RequestSamples(Sample *sample, const Scene *scene) {
		if (strategy == SAMPLE_ALL_UNIFORM) {
			// Allocate and request samples for sampling all lights
			u_int nLights = scene->lights.size();
			lightSampleOffset = new int[nLights];
			bsdfSampleOffset = new int[nLights];
			bsdfComponentOffset = new int[nLights];
			for (u_int i = 0; i < nLights; ++i) {
				const Light *light = scene->lights[i];
				int lightSamples =
					scene->sampler->RoundSize(light->nSamples);
				lightSampleOffset[i] = sample->Add2D(lightSamples);
				bsdfSampleOffset[i] = sample->Add2D(lightSamples);
				bsdfComponentOffset[i] = sample->Add1D(lightSamples);
			}
			lightNumOffset = -1;
		}
		else {
			// Allocate and request samples for sampling one light
			lightSampleOffset = new int[1];
			lightSampleOffset[0] = sample->Add2D(1);
			lightNumOffset = sample->Add1D(1);
			bsdfSampleOffset = new int[1];
			bsdfSampleOffset[0] = sample->Add2D(1);
			bsdfComponentOffset = new int[1];
			bsdfComponentOffset[0] = sample->Add1D(1);
		}
	}
	
	// Aerial perspective methods
	float tau(float h0, float theta, float s, float T, int i, float lambda) const; //returns the extinction factor tau
	void spectrum_to_xyz(double *x, double *y, double *z) const;
	
private:
	// DirectLighting Private Data
	LightStrategy strategy;
	mutable int rayDepth; // NOBOOK
	int maxDepth; // NOBOOK
	// Declare sample parameters for light source sampling
	int *lightSampleOffset, lightNumOffset;
	int *bsdfSampleOffset, *bsdfComponentOffset;
	mutable float *avgY, *avgYsample, *cdf;
	mutable float overallAvgY;

	// Aerial perspective variables
	mutable spectrum41_t spectrum41;
	float turbidity;
};
// DirectLighting Method Definitions
DirectLighting::~DirectLighting() {
	delete[] avgY;
	delete[] avgYsample;
	delete[] cdf;
}
DirectLighting::DirectLighting(LightStrategy st, int md, const float &turb) {
	maxDepth = md;
	rayDepth = 0;
	strategy = st;
	avgY = avgYsample = cdf = NULL;
	overallAvgY = 0.;

	spectrum41.nsamples = 41;
	spectrum41.start_wavelength = 380;
	spectrum41.end_wavelength = 780;
	turbidity = turb;
}
Spectrum DirectLighting::Li(const Scene *scene,
		const RayDifferential &ray, const Sample *sample,
		float *alpha) const {
	Intersection isect;
	Spectrum L(0.);
	Point phit;
	if (scene->Intersect(ray, &isect)) {
		if (alpha) *alpha = 1.;
		// Evaluate BSDF at hit point
		BSDF *bsdf = isect.GetBSDF(ray);
		Vector wo = -ray.d;
		const Point &p = bsdf->dgShading.p;
		phit = bsdf->dgShading.p;
		const Normal &n = bsdf->dgShading.nn;
		// Compute emitted light if ray hit an area light source
		L += isect.Le(wo);
		// Compute direct lighting for _DirectLighting_ integrator
		if (scene->lights.size() > 0) {
			// Apply direct lighting strategy
			switch (strategy) {
				case SAMPLE_ALL_UNIFORM:
					L += UniformSampleAllLights(scene, p, n, wo, bsdf,
						sample, lightSampleOffset, bsdfSampleOffset,
						bsdfComponentOffset);
					break;
				case SAMPLE_ONE_UNIFORM:
					L += UniformSampleOneLight(scene, p, n, wo, bsdf,
						sample, lightSampleOffset[0], lightNumOffset,
						bsdfSampleOffset[0], bsdfComponentOffset[0]);
					break;
				case SAMPLE_ONE_WEIGHTED:
					L += WeightedSampleOneLight(scene, p, n, wo, bsdf,
						sample, lightSampleOffset[0], lightNumOffset,
						bsdfSampleOffset[0], bsdfComponentOffset[0], avgY,
						avgYsample, cdf, overallAvgY);
					break;
			}
		}
		if (rayDepth++ < maxDepth) {
			Vector wi;
			// Trace rays for specular reflection and refraction
			Spectrum f = bsdf->Sample_f(wo, &wi,
				BxDFType(BSDF_REFLECTION | BSDF_SPECULAR));
			if (!f.Black()) {
				// Compute ray differential _rd_ for specular reflection
				RayDifferential rd(p, wi);
				rd.hasDifferentials = true;
				rd.rx.o = p + isect.dg.dpdx;
				rd.ry.o = p + isect.dg.dpdy;
				// Compute differential reflected directions
				Normal dndx = bsdf->dgShading.dndu * bsdf->dgShading.dudx +
					bsdf->dgShading.dndv * bsdf->dgShading.dvdx;
				Normal dndy = bsdf->dgShading.dndu * bsdf->dgShading.dudy +
					bsdf->dgShading.dndv * bsdf->dgShading.dvdy;
				Vector dwodx = -ray.rx.d - wo, dwody = -ray.ry.d - wo;
				float dDNdx = Dot(dwodx, n) + Dot(wo, dndx);
				float dDNdy = Dot(dwody, n) + Dot(wo, dndy);
				rd.rx.d = wi -
				          dwodx + 2 * Vector(Dot(wo, n) * dndx +
						  dDNdx * n);
				rd.ry.d = wi -
				          dwody + 2 * Vector(Dot(wo, n) * dndy +
						  dDNdy * n);
				L += scene->Li(rd, sample) * f * AbsDot(wi, n);
			}
			f = bsdf->Sample_f(wo, &wi,
				BxDFType(BSDF_TRANSMISSION | BSDF_SPECULAR));
			if (!f.Black()) {
				// Compute ray differential _rd_ for specular transmission
				RayDifferential rd(p, wi);
				rd.hasDifferentials = true;
				rd.rx.o = p + isect.dg.dpdx;
				rd.ry.o = p + isect.dg.dpdy;
				
				float eta = bsdf->eta;
				Vector w = -wo;
				if (Dot(wo, n) < 0) eta = 1.f / eta;
				
				Normal dndx = bsdf->dgShading.dndu * bsdf->dgShading.dudx + bsdf->dgShading.dndv * bsdf->dgShading.dvdx;
				Normal dndy = bsdf->dgShading.dndu * bsdf->dgShading.dudy + bsdf->dgShading.dndv * bsdf->dgShading.dvdy;
				
				Vector dwodx = -ray.rx.d - wo, dwody = -ray.ry.d - wo;
				float dDNdx = Dot(dwodx, n) + Dot(wo, dndx);
				float dDNdy = Dot(dwody, n) + Dot(wo, dndy);
				
				float mu = eta * Dot(w, n) - Dot(wi, n);
				float dmudx = (eta - (eta*eta*Dot(w,n))/Dot(wi, n)) * dDNdx;
				float dmudy = (eta - (eta*eta*Dot(w,n))/Dot(wi, n)) * dDNdy;
				
				rd.rx.d = wi + eta * dwodx - Vector(mu * dndx + dmudx * n);
				rd.ry.d = wi + eta * dwody - Vector(mu * dndy + dmudy * n);
				L += scene->Li(rd, sample) * f * AbsDot(wi, n);
			}
		}
		--rayDepth;
	}
	else {
		// Handle ray with no intersection
		if (alpha) *alpha = 0.;
		for (u_int i = 0; i < scene->lights.size(); ++i)
			L += scene->lights[i]->Le(ray);
		if (alpha && !L.Black()) *alpha = 1.;
		return L;
	}
	// Return L;
	//float h0 = ray.o.y;
	//Vector up = Vector(0.f, 1.f, 0.f);
	//float theta = acos(Dot(ray.d, up));
	//Vector s_ = Vector(phit-ray.o);
	//float s = s_.Length();
	//float T = turbidity;
	//
	//int i;
	//float lambda;
	//for (i = 0, lambda = 380; lambda < 780.1; i++, lambda += 10) {
	//	spectrum41.samples[i] = tau(h0, theta, s, T, i, lambda);
	//}
	//double X, Y, Z;
	//spectrum_to_xyz(&X, &Y, &Z);
	//float rgb[3];
	//rgb[0] =  3.240479f * X + -1.537150f * Y + -0.498535f * Z;
	//rgb[1] =  -0.969256f * X +  1.875991f * Y +  0.041556f * Z;
	//rgb[2] =  0.055648f * X + -0.204043f * Y +  1.057311f * Z;
	//
	//spectrum41.r = rgb[0];
	//spectrum41.g = rgb[1];
	//spectrum41.b = rgb[2];

	//return L * Spectrum(rgb);
	return L;
}

inline float EvalFunc(float B, float x)
{
    if (fabs(B*x)< 0.01) return x;
    return (1-exp(-B*x))/B;
}

float DirectLighting::tau(float h0, float theta, float s, float T, int i, float lambda) const {
	float n = 1.0003; // refreactivity of air
	float N = 2.545e25; // number of molecules per unit volume
	float pn = 0.035; // depolarization factor
	lambda = lambda * 1e-3;
	float lambda4 = lambda*lambda*lambda*lambda;

	// Rayleigh scattering
	float beta_m = ((8*pow(M_PI,3)*((n*n-1)*(n*n-1))) / (3*N*lambda4)) * ((6+3*pn) / (6-7*pn));
	beta_m = beta_m * (1+(cos(theta)*cos(theta)));

	// Mie scattering
	float c = (0.6544*T - 0.6510)*1e-16; // concentration factor, varies with turbidity (T)
	int v = 4; // Junge's exponent
	float beta_p = 0.434*c*M_PI * pow(((2*M_PI)/lambda),v-2)*K[i];

	// calculate tau
	float costheta = cos(theta);
    float B_1 = Alpha_1 * costheta;
    float B_2 = Alpha_2 * costheta;
    float constTerm_1 = exp(-Alpha_1*h0) * EvalFunc(B_1, s);
    float constTerm_2 = exp(-Alpha_2*h0) * EvalFunc(B_2, s);
 
	// return tau (extinction factor)
    return (exp(-beta_p * constTerm_1) *
	    exp(-beta_m * constTerm_2));
}

void DirectLighting::spectrum_to_xyz(double *x, double *y, double *z) const {
    int i;
    double lambda, X = 0, Y = 0, Z = 0, XYZ;

    static double cie_colour_match[81][3] = {
        {0.0014,0.0000,0.0065}, {0.0022,0.0001,0.0105}, {0.0042,0.0001,0.0201},
        {0.0076,0.0002,0.0362}, {0.0143,0.0004,0.0679}, {0.0232,0.0006,0.1102},
        {0.0435,0.0012,0.2074}, {0.0776,0.0022,0.3713}, {0.1344,0.0040,0.6456},
        {0.2148,0.0073,1.0391}, {0.2839,0.0116,1.3856}, {0.3285,0.0168,1.6230},
        {0.3483,0.0230,1.7471}, {0.3481,0.0298,1.7826}, {0.3362,0.0380,1.7721},
        {0.3187,0.0480,1.7441}, {0.2908,0.0600,1.6692}, {0.2511,0.0739,1.5281},
        {0.1954,0.0910,1.2876}, {0.1421,0.1126,1.0419}, {0.0956,0.1390,0.8130},
        {0.0580,0.1693,0.6162}, {0.0320,0.2080,0.4652}, {0.0147,0.2586,0.3533},
        {0.0049,0.3230,0.2720}, {0.0024,0.4073,0.2123}, {0.0093,0.5030,0.1582},
        {0.0291,0.6082,0.1117}, {0.0633,0.7100,0.0782}, {0.1096,0.7932,0.0573},
        {0.1655,0.8620,0.0422}, {0.2257,0.9149,0.0298}, {0.2904,0.9540,0.0203},
        {0.3597,0.9803,0.0134}, {0.4334,0.9950,0.0087}, {0.5121,1.0000,0.0057},
        {0.5945,0.9950,0.0039}, {0.6784,0.9786,0.0027}, {0.7621,0.9520,0.0021},
        {0.8425,0.9154,0.0018}, {0.9163,0.8700,0.0017}, {0.9786,0.8163,0.0014},
        {1.0263,0.7570,0.0011}, {1.0567,0.6949,0.0010}, {1.0622,0.6310,0.0008},
        {1.0456,0.5668,0.0006}, {1.0026,0.5030,0.0003}, {0.9384,0.4412,0.0002},
        {0.8544,0.3810,0.0002}, {0.7514,0.3210,0.0001}, {0.6424,0.2650,0.0000},
        {0.5419,0.2170,0.0000}, {0.4479,0.1750,0.0000}, {0.3608,0.1382,0.0000},
        {0.2835,0.1070,0.0000}, {0.2187,0.0816,0.0000}, {0.1649,0.0610,0.0000},
        {0.1212,0.0446,0.0000}, {0.0874,0.0320,0.0000}, {0.0636,0.0232,0.0000},
        {0.0468,0.0170,0.0000}, {0.0329,0.0119,0.0000}, {0.0227,0.0082,0.0000},
        {0.0158,0.0057,0.0000}, {0.0114,0.0041,0.0000}, {0.0081,0.0029,0.0000},
        {0.0058,0.0021,0.0000}, {0.0041,0.0015,0.0000}, {0.0029,0.0010,0.0000},
        {0.0020,0.0007,0.0000}, {0.0014,0.0005,0.0000}, {0.0010,0.0004,0.0000},
        {0.0007,0.0002,0.0000}, {0.0005,0.0002,0.0000}, {0.0003,0.0001,0.0000},
        {0.0002,0.0001,0.0000}, {0.0002,0.0001,0.0000}, {0.0001,0.0000,0.0000},
        {0.0001,0.0000,0.0000}, {0.0001,0.0000,0.0000}, {0.0000,0.0000,0.0000}
    };

    for (i = 0, lambda = 380; lambda < 780.1; i+=2, lambda += 10) {

        X += spectrum41.samples[i / 2] * cie_colour_match[i][0];
        Y += spectrum41.samples[i / 2] * cie_colour_match[i][1];
        Z += spectrum41.samples[i / 2] * cie_colour_match[i][2];

    }

	XYZ = (X + Y + Z);
	if (fabs(XYZ) > 1.0e-48) { 
        *x = X / XYZ;
        *y = Y / XYZ;
        *z = Z / XYZ;
    } else {
        *x = X;
        *y = Y;
        *z = Z;
	}
}

extern "C" DLLEXPORT SurfaceIntegrator *CreateSurfaceIntegrator(const ParamSet &params) {
	int maxDepth = params.FindOneInt("maxdepth", 5);
	LightStrategy strategy;
	string st = params.FindOneString("strategy", "all");
	if (st == "one") strategy = SAMPLE_ONE_UNIFORM;
	else if (st == "all") strategy = SAMPLE_ALL_UNIFORM;
	else if (st == "weighted") strategy = SAMPLE_ONE_WEIGHTED;
	else {
		Warning("Strategy \"%s\" for direct lighting unknown. "
			"Using \"all\".", st.c_str());
		strategy = SAMPLE_ALL_UNIFORM;
	}

	// Aerial perpective input
	float turb = params.FindOneFloat("turbidity", 4.0f); 

	return new DirectLighting(strategy, maxDepth, turb);
}
